//////////////////////////////////////////////////////////////////////////////
#include "stdlib.h"
#include "stdio.h"
#include "conio.h"
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
#define NUMTOKENS 12
#define MAXTRACESIZE 1000
#define STACKSIZE 100
#define MISSLATENCY 4
#define EXIT -1
#define BRANCH 1
#define STORE 2
#define LOAD 3
#define ALU 4
#define SW 2
#define LW 3
#define BNEZ 5
#define BNE 6
#define ADD 7
#define SUB 8
#define MUL 9
#define DIV 10
#define FADD  11
#define FSUB 12
#define FMUL 13
#define FDIV 14
#define DADDUI 15
#define INTREGS 32
#define FPREGS 32
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
int  ISSUESIZE; 
int  COMMITSIZE;
int  ROBSIZE; 
int  NUMCDB; 
int  LSQSIZE; 
int  SCACHESIZE; 
int  NUMRSBU; 
int  NUMRSINTALU; 
int  NUMRSMULT; 
int  MULTSTAGES; 
int  NUMRSDIV;
int  DIVCYCLES; 
int  NUMRSFPADD; 
int  FPADDSTAGES;
int  MISSPROB; 
int  MPPROB;
int  IC=0;
int  fetchstallcycles=0;
int  issuestallcycles=0;
int  intALUstallcycles=0;
int	 multUnitstallcycles=0;
int  intALUutil=0;
int  multUnitutil=0;
static int numissued[100];
static int numcommitted[100];
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
int readConfigFile(void);
int readTraceFile(char *traceFile);
void parseLine(char *line, int *tokens);
void readTokens(int *tokens, int i);
int commit(void);
void resetState(void);
void initializeREGS();
void updateReservationStations(void);
void writeback(void);
void execute(void);
void issue(void);
void fetch(int branchMiss);
int getBranchPrediction();
int getLoadPrediction();
void assignValue(int n,int a,int b,int c,int d,int e,int f,int g,int h,int i,int j,int k,int l);
void printInstrInfo(int c);
void outputUtils(void);
int parseParam(char *s);
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
typedef struct traceElementType
{
	int addr;
    int opcode;
    int dest;
    int src1;
    int src2;
    int imm;
    int opd1;
    int opd2;
    int branchDest;
    int takenCount;
    int bPredict;
    int bOutcome;
    int loadMiss;
}traceElement;
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
traceElement trace[MAXTRACESIZE];
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
typedef struct RegisterType
{
   int data;
   int tag;
   int valid;
}Register;
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Register REGS[INTREGS+FPREGS];
//////////////////////////////////////////////////////////////////////////////



//////////////////////////////////////////////////////////////////////////////
class commonDataBus
{
public:
   int *data;//[NUMCDB];
   int *tag;//[NUMCDB];
   int *util;
   
   int commonDataBusInit()
   {
	   if((data=(int*)malloc(sizeof(int)*NUMCDB))==NULL)
	   {
		   printf("\nNot enough memory to allocate CDB.data array");
		   return 0;
	   }

	   if((tag=(int*)malloc(sizeof(int)*NUMCDB))==NULL)
	   {
		   printf("\nNot enough memory to allocate CDB.tag array");
		   return 0;
	   }
	   if((util=(int*)calloc(NUMCDB,sizeof(int)))==NULL)
	   {
		   printf("\nNot enough memory to allocate CDB.util array");
		   return 0;
	   }
	   

    for(int i=0;i<NUMCDB;i++)
         tag[i]=-1;
		return 1;
   }
  

   int checkTag(int TAG)
   {
      for(int i=0;i<NUMCDB;i++)
         if(tag[i]==TAG)
            return 1;
      return 0;
   }
   
   int read(int TAG) 
   {
      for(int i=0;i<NUMCDB;i++)
         if(tag[i]==TAG)
            return data[i];
      return -1;  
   }
   
   void write(int DATA, int NUM)
   {
      data[NUM-1]=DATA;
   }
   
   void updateTag(int TAG, int NUM)
   {
      tag[NUM-1]=TAG;
	  util[NUM-1]++;
   }

   void floatTag(int i)
   {
	 //for(int i=0;i<NUMCDB;i++)
         //if(tag[i]!=-1)
     tag[i]=-1;
   }
   
   int readTag(int NUM)
   {
      return tag[NUM-1];
   }
   
   void arbiter(int vBRU,int vDIV,int vLSQ,int vMULT,int vFPADD,int vALU);
	
   void display(void)
   {
	   for(int i=0; i<NUMCDB; i++)
		printf("\nCDB[%d]->TAG:%d",i,tag[i]);
	   printf("\n*****************************");
   }
};
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
commonDataBus CDB;
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
typedef struct ROBslotType
{
   int instrType; //any one of four categories:BRANCH,ALU,LOAD,STORE
   int dest;   // what about floating point and integer regs separation? Solved
   int data;   //the value to write 
   int ready;  //slot is ready for commit
   int tag;    //for stores
   int IP; //trace pointer for printing out info each cycle
   int dest_v; //for stores
   int data_v;
}ROBslot;
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
class reOrderBuffer
{
public:
   int head;
   int tail;
   int count;
   int lastIndex;

   int *ROBocc;
   ROBslot *R;//[ROBSIZE];

   int reOrderBufferAlloc(void)
   {
      // the constructor for initialization tasks
	   if((R=(ROBslot*)malloc(sizeof(ROBslot)*ROBSIZE))==NULL)
	   {
		   printf("\nNot enough memory to allocate ROB data array");
		   return 0;
	   }
	   if((ROBocc=(int*)calloc(ROBSIZE+1,sizeof(int)))==NULL)
	   {
		   printf("\nNot enough memory to allocate ROB occupancy array");
		   return 0;
	   }

      head=0;
      tail=0;
      count=0;
      lastIndex=-1;
	  int i;
	  for(i=0;i<ROBSIZE;i++)
	  {  
		  R[i].tag=-1;
		  R[i].ready=0;
		  R[i].instrType=-1;
		  R[i].data=0;
		  R[i].dest=-1;
		  R[i].IP=-1;
		  R[i].dest_v=0;
		  R[i].data_v=0;
	  }
	  return 1;
   }
   
   int full(void)
   {
      if(count==ROBSIZE)
         return 1;
      else
         return 0;
   }
   
   int empty(void)
   {
      if(count==0)
         return 1;
      else
         return 0;
   }
   
   int enqueue(int Opcode,int pointer)
   {  
      if(full())
         return -1;
      
      if(Opcode==BNEZ || Opcode==BNE)
         R[tail].instrType = BRANCH;
      else if(Opcode==SW)
         R[tail].instrType = STORE;
      else if(Opcode==LW)
         R[tail].instrType = LOAD;
      else
         R[tail].instrType = ALU;
	  R[tail].IP=pointer;
      count++;
      lastIndex=tail;
      tail++;
      if(tail==ROBSIZE)
       tail=0;      
      return lastIndex;      
   }
   
   int dequeue(void)
   {
      if(empty())
         return -1;
      
      count--;
      
      if(head==ROBSIZE-1)
       head=0;
      else
       head++;
      
      return head;
   }   
   
   int instrType_r(int x)
   {
      return R[x].instrType;
   }
   
   int dest_r(int x)
   {
      return R[x].dest;
   }
    
   int data_r(int x)
   {
      return R[x].data;
   }
   
   int ready_r(int x)
   {
      return R[x].ready;
   }
   
   int tag_r(int x)
   {
      return R[x].tag;
   }
   
   int IP_r(int x)
   {
      return R[x].IP;
   }
	
   int data_v_r(int x)
   {
      return R[x].data_v;
   }

   void instrType_w(int x, int a)
   {
      R[x].instrType=a;
   }
   
   void dest_w(int x, int a)
   {
      R[x].dest=a;
   }
    
   void data_w(int x, int a)
   {
      R[x].data=a;
   }

   void data_v_w(int x, int a)
   {
      R[x].data_v=a;
   }
   
   void ready_w(int x, int a)
   {
      R[x].ready=a;
   }

   void dest_v_w(int x, int a)
   {
      R[x].dest_v=a;
   }
   
   void tag_w(int x, int a)
   {
      R[x].tag=a;
   }
   
   void reset()
   {
      head=0;
      tail=0;
      count=0;
      lastIndex=-1;
	  for(int i=0;i<ROBSIZE;i++)
	  {  
		  R[i].tag=-1;
		  R[i].ready=0;
		  R[i].instrType=-1;
		  R[i].data=0;
		  R[i].dest=-1;
		  R[i].IP=-1;
		  R[i].dest_v=0;
		  R[i].data_v=0;
	  }
   }

	void calcUtil(void)
	{
		ROBocc[count]++;
	}

   void updateROB(void)
   {
	   int i;
	   int n;

		for(int j=1;j<=NUMCDB;j++)
        {
		  i=head;
		  n=count;
		  if(CDB.readTag(j)!=-1)
		  {	
			printf("\nUpdating ROB[%d]:  ",CDB.readTag(j));
			printInstrInfo(R[CDB.readTag(j)].IP);
			R[CDB.readTag(j)].data=CDB.read(CDB.readTag(j));
			R[CDB.readTag(j)].ready=1;
			printf("\nChecking for STORE instructions updates pending in ROB...");
			while(n>0)
			{
				if(R[i].tag == CDB.readTag(j) && R[i].instrType==STORE)
				{
					printf("\nUPDATING %d:",i);
					printInstrInfo(R[i].IP);
					R[i].data_v=1;
					if(R[i].dest_v==1)
						R[i].ready=1;
				}
				n--;
				if(i==ROBSIZE-1)
					i=0;
				else
					i++;
			}
		  }
        }
   }

   void display(void)
   {
	   int i=head;
	   int n=count;
	   
	   printf("\n**********ROB CONTENTS**********");
	   if(count==0)
		   printf("\n!!!!!!!!!EMPTY!!!!!!!!");
	   while(n>0)
	   {
		   if(R[i].ready)
		   {
			   printf("\n%d:READY ",i);
			   printInstrInfo(R[i].IP);
		   }
		   else
		   {
			   printf("\n%d:NOT READY ",i);
			   printInstrInfo(R[i].IP);
		   }
		n--;
      
		if(i==ROBSIZE-1)
			i=0;
		else
			i++;
	   }
	   printf("\n********************************");
   }
};
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
reOrderBuffer ROB;
//////////////////////////////////////////////////////////////////////////////



//////////////////////////////////////////////////////////////////////////////
class RSslotType1
{
   public:
   int full;
   int op;
   int tag;
   Register opd1, opd2;
   int ip;
};
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
class RSslotType2
{
   public:
   int full;
   int op;
   int tag;
   Register opd1;
   int imm;
   int ip;
};
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
class simpleALU
{
	public:
   int op;
   int opd1, opd2;
   int result;
   int tagOut;
   int outValid;
   int stall;
   int ack;
   int reqCDB;
   
   RSslotType1 *RS;//[NUMRSINTALU];
   int *RSocc;
   
   int simpleALUAlloc()
   {
	   if((RS=(RSslotType1*)malloc(sizeof(RSslotType1)*NUMRSINTALU))==NULL)
	   {
		   printf("\nNot enough memory to allocate RS data array for INT ALU");
		   return 0;
	   }
	   if((RSocc=(int*)calloc(NUMRSINTALU+1,sizeof(int)))==NULL)
	   {
		   printf("\nNot enough memory to allocate INT ALU RS occupancy array");
		   return 0;
	   }
	  stall=0;
      ack=0;
      outValid=0;
	  tagOut=-1;
	  reqCDB=0;
      
      for(int i=0;i<NUMRSINTALU;i++)
      {
         RS[i].full=0;
      }
	  return 1;
   }
  
   	void calcUtil(void)
	{
		int full=0;
	   for(int i=0;i<NUMRSINTALU;i++)
	   {
			if(RS[i].full)
				full++;
	   }
		RSocc[full]++;
	}

   void displayRS(void)
   {
	   int full=0;
	   printf("\n****INTEGER ALU RS CONTENTS****");
	   for(int i=0;i<NUMRSINTALU;i++)
	   {
			if(RS[i].full)
			{
				if(RS[i].opd1.valid && RS[i].opd2.valid)
				{	
					printf("\nREADY: ");
					printInstrInfo(RS[i].ip);
				}
				else
				{
					printf("\nWAITING: ");
					printInstrInfo(RS[i].ip);
					if(!RS[i].opd1.valid)
						printf(" OPD1 ");
					if(!RS[i].opd2.valid)
						printf(" OPD2 ");
				}
				full++;
			}
			
	   }
	   if(full==0)
		   printf("\n!!!!!!!!!!{EMPTY}!!!!!!!!!!");
	   printf("\n*******************************");
   }
	   
   int getFreeRS(void)
   {
     for(int i=0;i<NUMRSINTALU;i++)
     {
      if(!RS[i].full)
         return i;
     }
     return -1;
   }
	
   void reset(void)
   {
      stall=0;
      ack=0;
      outValid=0;
      
      for(int i=0;i<NUMRSINTALU;i++)
      {
         RS[i].full=0;
      }
   }
   
   int issue(traceElement i)
   {
	  int y;
      int x=getFreeRS();
      if(x==-1)
         return 1;
      
      y=ROB.enqueue(i.opcode,i.addr);   
      RS[x].op=i.opcode;
      RS[x].full=1;
      RS[x].tag=y;
	  RS[x].ip=i.addr;
      
      //for instruction i source operand 1
      if(i.src1!=-1)
      {
         if(REGS[i.src1].valid)
         {
            RS[x].opd1.valid=1;
            RS[x].opd1.data=REGS[i.src1].data;
         }
         else if(ROB.ready_r(REGS[i.src1].tag))
         {
            RS[x].opd1.valid=1;
            RS[x].opd1.data=ROB.data_r(REGS[i.src1].tag);
         }
         else if(CDB.checkTag(REGS[i.src1].tag))
         {
            RS[x].opd1.valid=1;
            RS[x].opd1.data=CDB.read(REGS[i.src1].tag);
		 }
         else
         {
            RS[x].opd1.valid=0;
            RS[x].opd1.tag=REGS[i.src1].tag;
         }
      }
      else
      {
          RS[x].opd1.valid=1;
          RS[x].opd1.data=i.opd1;
      }   
      
      //for instruction i source operand 2
      if(i.src2!=-1)
      {
         if(REGS[i.src2].valid)
         {
            RS[x].opd2.valid=1;
            RS[x].opd2.data=REGS[i.src2].data;
         }
         else if(ROB.ready_r(REGS[i.src2].tag))
         {
            RS[x].opd2.valid=1;
            RS[x].opd2.data=ROB.data_r(REGS[i.src2].tag);
         }
         else if(CDB.checkTag(REGS[i.src2].tag))
         {
            RS[x].opd2.valid=1;
            RS[x].opd2.data=CDB.read(REGS[i.src2].tag);
		 }
         else
         {
            RS[x].opd2.valid=0;
            RS[x].opd2.tag=REGS[i.src2].tag;
         }
      }
      else
      {
          RS[x].opd2.valid=1;
          RS[x].opd2.data=i.opd2;
      }   
            
       //for instruction i destination operand
      REGS[i.dest].tag=y;
	  REGS[i.dest].valid=0;
      ROB.dest_w(y,i.dest);
      ROB.ready_w(y,0);
	  return 0;
   }
   
   int getReadyRS(void)
   {
     for(int i=0; i<NUMRSINTALU; i++)
     {
	  if(RS[i].full)
		if(RS[i].opd1.valid && RS[i].opd2.valid)
         return i;
     }
     return -1;
   }
      
   void BeginEx(int x)
   {
	  intALUutil++;
      outValid=1;   // who will reset this? the arbiter?
      result=RS[x].opd1.data + RS[x].opd2.data; //can be generic by switch(op)
      tagOut=RS[x].tag;
	  reqCDB=1;
   }
   
	
   void freeRS(int x)
   {
      RS[x].full=0;
      RS[x].tag=-1;
   }
   
   void check4Stall(void)
   {
      if(reqCDB==1)
         stall=1;
   }
   
   void execute(void)
   {
      check4Stall();
	  int readyRS=getReadyRS();
	  printf("\n********INT ALU EXECUTION***********");
      if(!stall)
      {
         if(readyRS!=-1)
         {
			printf("\nEXECUTING: ");
			printInstrInfo(RS[readyRS].ip);
            BeginEx(readyRS);
            freeRS(readyRS);
         }
      }
	  else
	  {
			if(readyRS!=-1)
			{
				//printf("\n!!!!!!!!!!INT ALU EXECUTE STALL CYCLE!!!!!!!!!!!!");
				printf("\nPENDING EXECUTION: ");
				printInstrInfo(RS[readyRS].ip);
				//intALUstallcycles++;
			}
			
	  }
	  displayEx();
	  printf("\n************************************");
   }
   
   void displayEx()
   {
	   printf("\nTag=%d:Valid=%d:Ack=%d:Stall=%d",tagOut,outValid,ack,stall);
   }

   void writeback(void)
   {
      if(ack)
      {
		 printf("\nResult on CDB ");
		 printInstrInfo(ROB.IP_r(tagOut));
         CDB.write(result,ack);
         CDB.updateTag(tagOut,ack);
         outValid=0;             // is this the proper place to do this?
//         ROB.ready_w(CDB.readTag(ack),1);
//         ROB.data_w(CDB.readTag(ack),CDB.read(CDB.readTag(ack)));
         ack=0;
		 reqCDB=0;
         stall=0;
      }
	  //else
	  //	   CDB.floatTag();

	  if(reqCDB==1)
	  {
			printf("\n!!!!!!!!!!INT ALU EXECUTE STALL CYCLE!!!!!!!!!!!!");
			printf("\nWAITING FOR CDB: ");
			printInstrInfo(ROB.IP_r(tagOut));
			intALUstallcycles++;
	  }

      //snooping function
      //updateRS();
   }
   
   void updateRS(void)
   {
     for(int i=0; i<NUMRSINTALU; i++)
     {
      if(RS[i].full)
      {
         for(int j=1;j<=NUMCDB;j++)
         {
            if(!RS[i].opd1.valid && RS[i].opd1.tag==CDB.readTag(j))
            {
			   printf("\nUpdating INT ALU RS[%d]: (OPD1) ",i);
			   printInstrInfo(RS[i].ip);
               RS[i].opd1.data=CDB.read(CDB.readTag(j));
               RS[i].opd1.valid=1;
            }
            if(!RS[i].opd2.valid && RS[i].opd2.tag==CDB.readTag(j))
            {
			   printf("\nUpdating INT ALU RS[%d]: (OPD2) ",i);
			   printInstrInfo(RS[i].ip);
               RS[i].opd2.data=CDB.read(CDB.readTag(j));
               RS[i].opd2.valid=1;
            }
         }
      }
     }
     
   }
   
 
};
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
simpleALU intALU;
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
class multCDBreqQueue
{
   public:
   int *buffer;
   
   int head;
   int tail;
   int count;
   int lastIndex;
   
   int multCDBreqQueueAlloc()
   {
	   if((buffer=(int*)malloc(sizeof(int)*MULTSTAGES))==NULL)
	   {
		   printf("\nNot enough memory to allocate MULT CDB req Q data array");
		   return 0;
	   }

      head=0;
      tail=0;
      count=0;
      lastIndex=0;
	  return 0;
   }

	void reset(void)
	{
	  head=0;
      tail=0;
      count=0;
      lastIndex=0;
	}

   int full()
   {
      if(count==MULTSTAGES)
         return 1;
      else
         return 0;      
   }
   
   int empty()
   {
      if(count==0)
         return 1;
      else
         return 0;      
   }
   
   int enqueue()
   {
      if(full())
         return -1;

      buffer[tail]=0; //enter a new counter value
      
      count++;
      lastIndex=tail;
      tail++;
      if(tail==MULTSTAGES)
       tail=0;      
      return lastIndex;      
   }
   
   int dequeue()
   {
      if(empty())
         return -1;
      
      count--;
      
      if(head==MULTSTAGES-1)
       head=0;
      else
       head++;
      
      return head;      
   }
   
   int readElement()
   {
      return buffer[head];
   }

   void increment()
   {
	   int i=head;
	   int n=count;
	   //printf("\n*****MULT CDB REQ Q CONTENTS*****");
	   if(count==0)
		   return;
		//   printf("\n!!!!!! {EMPTY} !!!!!!");
	   else
	   {
		while(n>0)
		{
		   	//printf("\n%d",buffer[i]);
			//printInstrInfo(buffer[i].addr);
			buffer[i]++;
			if(i==MULTSTAGES-1)
				i=0;
			else
				i++;
			n--;
		}
	   }
	   //printf("\nTotal Queued:%d",count);
	   //printf("\n*********************");
   }

   void display()
   {
	   int i=head;
	   int n=count;
	   printf("\n*****MULT CDB REQ Q CONTENTS*****\n");
	   if(count==0)
		   printf("!!!!!! {EMPTY} !!!!!!");
	   else
	   {
		while(n>0)
		{
		   	printf("Q[%d]=%d ",i,buffer[i]);
			//printInstrInfo(buffer[i].addr);
			if(i==MULTSTAGES-1)
				i=0;
			else
				i++;
			n--;
		}
	   }
	   printf("\nTotal Queued:%d",count);
	   printf("\n*********************");
   }
   
};
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
class pipelinedMultipler
{
public:
   int op;
   int opd1, opd2;
   int result;
   int tagOut;
   int outValid;
   int *iResult;//[MULTSTAGES];
   int *iTag;//[MULTSTAGES];
   int *iOutValid;//[MULTSTAGES];
   int stall;
   int reqCDB;
   int ack;
   multCDBreqQueue CDBreqQ;

   RSslotType1 *RS;//[MAXRSMULT];
   int *RSocc;
   
   int pipelinedMultiplierAlloc()
   {
	   CDBreqQ.multCDBreqQueueAlloc();
	   if((RS=(RSslotType1*)malloc(sizeof(RSslotType1)*NUMRSMULT))==NULL)
	   {
		   printf("\nNot enough memory to allocate RS data array for MULT");
		   return 0;
	   }
	   if((RSocc=(int*)calloc(NUMRSMULT+1,sizeof(int)))==NULL)
	   {
		   printf("\nNot enough memory to allocate MULT RS occupancy array");
		   return 0;
	   }
	   if(MULTSTAGES>1)
	   {
		    if((iResult=(int*)calloc(MULTSTAGES-1,sizeof(int)))==NULL)
			{
				printf("\nNot enough memory to allocate MULT iResult array");
				return 0;
			}
			if((iTag=(int*)calloc(MULTSTAGES-1,sizeof(int)))==NULL)
			{
				printf("\nNot enough memory to allocate MULT iTag array");
				return 0;
			}
			if((iOutValid=(int*)calloc(MULTSTAGES-1,sizeof(int)))==NULL)
			{
				printf("\nNot enough memory to allocate MULT iOutValid array");
				return 0;
			}
	   }
	  stall=0;
      ack=0;
      outValid=0;
	  tagOut=-1;
	  reqCDB=0;
      
      for(int i=0;i<NUMRSMULT;i++)
      {
         RS[i].full=0;
      }
	  for(i=0;i<MULTSTAGES-1;i++)
      {
         iTag[i]=-1;
      }
	  return 1;
   }
  
   	void calcUtil(void)
	{
		int full=0;
	   for(int i=0;i<NUMRSMULT;i++)
	   {
			if(RS[i].full)
				full++;
	   }
		RSocc[full]++;
	}


	void displayRS(void)
    {
	   int full=0;
	   printf("\n****MULT RS CONTENTS****");
	   for(int i=0;i<NUMRSMULT;i++)
	   {
			if(RS[i].full)
			{
				if(RS[i].opd1.valid && RS[i].opd2.valid)
				{	
					printf("\nREADY: ");
					printInstrInfo(RS[i].ip);
				}
				else
				{
					printf("\nWAITING: ");
					printInstrInfo(RS[i].ip);
					if(!RS[i].opd1.valid)
						printf(" OPD1 ");
					if(!RS[i].opd2.valid)
						printf(" OPD2 ");
				}
				full++;
			}
			
	   }
	   if(full==0)
		   printf("\n!!!!!!!!!!{EMPTY}!!!!!!!!!!");
	   printf("\n*******************************");
   }


   int getFreeRS(void)
   {
     for(int i=0; i<NUMRSMULT; i++)
     {
      if(!RS[i].full)
         return i;
     }
     return -1;
   }
   

   void reset(void)
   {
      stall=0;
      ack=0;
      outValid=0;
      reqCDB=0;
	  CDBreqQ.reset();	  
      for(int i=0;i<NUMRSMULT;i++)
      {
         RS[i].full=0;
      }
   }

   int issue(traceElement i)
   {
	  int y;
      int x=getFreeRS();
      if(x==-1)
         return 1;
      
      y=ROB.enqueue(i.opcode,i.addr);   
      RS[x].op=i.opcode;
      RS[x].full=1;
      RS[x].tag=y;
	  RS[x].ip=i.addr;
	  //for instruction i source operand 1
      if(i.src1!=-1)
      {
         if(REGS[i.src1].valid)
         {
            RS[x].opd1.valid=1;
            RS[x].opd1.data=REGS[i.src1].data;
         }
         else if(ROB.ready_r(REGS[i.src1].tag))
         {
            RS[x].opd1.valid=1;
            RS[x].opd1.data=ROB.data_r(REGS[i.src1].tag);
         }
         else if(CDB.checkTag(REGS[i.src1].tag))
         {
            RS[x].opd1.valid=1;
            RS[x].opd1.data=CDB.read(REGS[i.src1].tag);
		 }
         else
         {
            RS[x].opd1.valid=0;
            RS[x].opd1.tag=REGS[i.src1].tag;
         }
      }
      else
      {
          RS[x].opd1.valid=1;
          RS[x].opd1.data=i.opd1;
      }   
      
      //for instruction i source operand 2
      if(i.src2!=-1)
      {
         if(REGS[i.src2].valid)
         {
            RS[x].opd2.valid=1;
            RS[x].opd2.data=REGS[i.src2].data;
         }
         else if(ROB.ready_r(REGS[i.src2].tag))
         {
            RS[x].opd2.valid=1;
            RS[x].opd2.data=ROB.data_r(REGS[i.src2].tag);
         }
         else if(CDB.checkTag(REGS[i.src2].tag))
         {
            RS[x].opd2.valid=1;
            RS[x].opd2.data=CDB.read(REGS[i.src2].tag);
		 }
         else
         {
            RS[x].opd2.valid=0;
            RS[x].opd2.tag=REGS[i.src2].tag;
         }
      }
      else
      {
          RS[x].opd2.valid=1;
          RS[x].opd2.data=i.opd2;
      }   
            
       //for instruction i destination operand
      REGS[i.dest].tag=y;
	  REGS[i.dest].valid=0;
      ROB.dest_w(y,i.dest);
      ROB.ready_w(y,0);
	  return 0;
   }
   
   int getReadyRS(void)
   {
     for(int i=0; i<NUMRSMULT; i++)
     {
	  if(RS[i].full)
		if(RS[i].opd1.valid && RS[i].opd2.valid)
         return i;
     }
     return -1;
   }
   
   void check4Stall(void)
   {
      if(reqCDB==1)
         stall=1;
   }
      
   void BeginEx(int x)
   {
	  multUnitutil++;
      if(MULTSTAGES>1)
      {
         iResult[0]=RS[x].opd1.data + RS[x].opd2.data; //can be generic by switch(op)
         iTag[0]=RS[x].tag; 
         iOutValid[0]=1;   // who will reset this? the arbiter?
		 CDBreqQ.enqueue();
      }
      else
      {
         result=RS[x].opd1.data + RS[x].opd2.data; //can be generic by switch(op)
         tagOut=RS[x].tag; 
         outValid=1;   // who will reset this? the arbiter?  
		 reqCDB=1;
      }
   }
   
   void executePipe(void)
   {
      if(MULTSTAGES>1)
      {
         outValid=iOutValid[MULTSTAGES-2];
         tagOut=iTag[MULTSTAGES-2];
         result=iResult[MULTSTAGES-2];
      
         for(int j=2;j<MULTSTAGES;j++)
         {
            iOutValid[MULTSTAGES-j]=iOutValid[MULTSTAGES-j-1];
            iResult[MULTSTAGES-j]=iResult[MULTSTAGES-j-1];
            iTag[MULTSTAGES-j]=iTag[MULTSTAGES-j-1];
         }
         iResult[0]=0;
         iTag[0]=-1; 
         iOutValid[0]=0;
		 CDBreqQ.increment();
		 printf("\n");
		 for(int i=0;i<MULTSTAGES-1;i++)
		 {
			printf("Tag[%d]=%d ",i,iTag[i]);
		 }
		 CDBreqQ.display();
		 if(!CDBreqQ.empty())
		 {
			if(CDBreqQ.readElement()==MULTSTAGES-1)
			{
				reqCDB=1;
				CDBreqQ.dequeue();
			}
		 }
      }
   }
   
   void freeRS(int x)
   {
      RS[x].full=0;
      RS[x].tag=-1;
   }
      
   void execute(void)
   {
      check4Stall();
	  int readyRS=getReadyRS();
	  printf("\n********MULT ALU EXECUTION***********");
      if(!stall)
      {
		 executePipe();
         if(readyRS!=-1)
         {
			printf("\nEXECUTING: ");
			printInstrInfo(RS[readyRS].ip);
            BeginEx(readyRS);
            freeRS(readyRS);
         }
      }
	  else
	  {
			if(readyRS!=-1)
			{
				//printf("\n!!!!!!!!!!INT ALU EXECUTE STALL CYCLE!!!!!!!!!!!!");
				printf("\nPENDING EXECUTION: ");
				printInstrInfo(RS[readyRS].ip);
				//intALUstallcycles++;
			}
			
	  }
	  displayEx();
	  printf("\n************************************");
   }
   
   void displayEx()
   {
	   printf("\nTag=%d:Valid=%d:Ack=%d:Stall=%d",tagOut,outValid,ack,stall);
   }

   void writeback(void)
   {
      if(ack)
      {
		 printf("\nResult on CDB ");
		 printInstrInfo(ROB.IP_r(tagOut));
         CDB.write(result,ack);
         CDB.updateTag(tagOut,ack);
         outValid=0;             // is this the proper place to do this?
//         ROB.ready_w(CDB.readTag(ack),1);
//         ROB.data_w(CDB.readTag(ack),CDB.read(CDB.readTag(ack)));
         ack=0;
		 reqCDB=0;
         stall=0;
      }
	  //else
	//	   CDB.floatTag();

	  if(reqCDB==1)
	  {
			printf("\n!!!!!!!!!!MULT UNIT EXECUTE STALL CYCLE!!!!!!!!!!!!");
			printf("\nWAITING FOR CDB: ");
			printInstrInfo(ROB.IP_r(tagOut));
			multUnitstallcycles++;
	  }
   }
   
   void updateRS(void)
   {
     for(int i=0; i<NUMRSMULT; i++)
     {
      if(RS[i].full)
      {
         for(int j=1;j<=NUMCDB;j++)
         {
            if(!RS[i].opd1.valid && RS[i].opd1.tag==CDB.readTag(j))
            {
			   printf("\nUpdating MULT UNIT RS[%d]: (OPD1) ",i);
			   printInstrInfo(RS[i].ip);
               RS[i].opd1.data=CDB.read(CDB.readTag(j));
               RS[i].opd1.valid=1;
            }
            if(!RS[i].opd2.valid && RS[i].opd2.tag==CDB.readTag(j))
            {
			   printf("\nUpdating MULT UNIT RS[%d]: (OPD2) ",i);
			   printInstrInfo(RS[i].ip);
               RS[i].opd2.data=CDB.read(CDB.readTag(j));
               RS[i].opd2.valid=1;
            }
         }
      }
     }
     
   }
   
} multUnit;
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//pipelinedMultiplier multUnit;
//////////////////////////////////////////////////////////////////////////////

class memoryUnit
{
public:
	int responseTime;
	int busy;
	int tag;
	int ip;
	int hit;

	memoryUnit()
	{
		responseTime=0;
		busy=0;
		tag=-1;
		ip=-1;
		hit=0;
	}
	
	void init(int tagIn, int ipIn, int Hit)
	{
		tag=tagIn;
		ip=ipIn;
		hit=Hit;
	}

	int free(void)
	{
		return !busy;
	}
	void takeReq(int tagIn, int ipIn, int hit);
	void completeReq(void);
}mem;


class loadStoreQueue
{
public:
   //int inpValid;
   int opIn;
   int opd1In;
   int result;
   int tagOut;
   int tagIn;
   int outValid;
   int stall;
   int ack;
   int reqCDB;
   int opOut1, ipOut1, tagOut1, immIn, ipIn;
   
   RSslotType2 *RS;//[MAXRSLSQ];
   int *RSocc;
   int head;
   int tail;
   int count;

   int loadStoreQueueAlloc()
   {
	   if((RS=(RSslotType2*)calloc(LSQSIZE,sizeof(RSslotType2)))==NULL)
	   {
		   printf("\nNot enough memory to allocate LSQ RS data array");
		   return 0;
	   }
	   if((RSocc=(int*)calloc(LSQSIZE+1,sizeof(int)))==NULL)
	   {
		   printf("\nNot enough memory to allocate LSQ RS occupancy array");
		   return 0;
	   }
	  stall=0;
      ack=0;
      outValid=0;
	  tagOut=-1;
	  reqCDB=0;
	  head=0;
	  tail=0;
	  count=0;
	  opIn=-1;
	  opd1In=0;
	  result=0;
      tagOut=-1;
      tagIn=-1;
      outValid=0;

      opOut1=-1; 
	  ipOut1=-1; 
	  tagOut1=-1;
	  immIn=0;
	  ipIn=-1;
      
      for(int i=0;i<LSQSIZE;i++)
      {
        RS[i].tag=-1;
      }
	  return 1;
   }
   

   void calcUtil(void)
	{
		RSocc[count]++;
	}

	int full(void)
	{
		if(count==LSQSIZE)
			return 1;
		else
			return 0;
	}

	int empty(void)
	{
		if(count==0)
			return 1;
		else
			return 0;
	}

   int getFreeRS(void)
   {
     if(!full())
      return tail;
     else
      return -1;
   }
   
   int issue(traceElement i)
   {
      int y;
      int x=getFreeRS();
      if(x==-1)
         return 1;
	  
      y=ROB.enqueue(i.opcode,i.addr);   
	  ROB.ready_w(y,0);
      RS[x].op=i.opcode;
      RS[x].full=1;
      RS[x].tag=y;
	  RS[x].ip=i.addr;      
      RS[x].imm = i.imm;

      //for instruction i source operand 1
	  if(i.opcode==LOAD)
	  {
         if(REGS[i.src1].valid)
         {
            RS[x].opd1.valid=1;
            RS[x].opd1.data=REGS[i.src1].data;
         }
         else if(ROB.ready_r(REGS[i.src1].tag))
         {
            RS[x].opd1.valid=1;
            RS[x].opd1.data=ROB.data_r(REGS[i.src1].tag);
         }
         else if(CDB.checkTag(REGS[i.src1].tag))
         {
            RS[x].opd1.valid=1;
            RS[x].opd1.data=CDB.read(REGS[i.src1].tag);
		 }
         else
         {
            RS[x].opd1.valid=0;
            RS[x].opd1.tag=REGS[i.src1].tag;
         }
      }
	  else
	  {
         if(REGS[i.src2].valid)
         {
            RS[x].opd1.valid=1;
            RS[x].opd1.data=REGS[i.src2].data;
         }
         else if(ROB.ready_r(REGS[i.src2].tag))
         {
            RS[x].opd1.valid=1;
            RS[x].opd1.data=ROB.data_r(REGS[i.src2].tag);
         }
         else if(CDB.checkTag(REGS[i.src2].tag))
         {
            RS[x].opd1.valid=1;
            RS[x].opd1.data=CDB.read(REGS[i.src2].tag);
		 }
         else
         {
            RS[x].opd1.valid=0;
            RS[x].opd1.tag=REGS[i.src2].tag;
         }
	  }
      //for instruction i destination operand
      if(i.opcode==LOAD)
      {
		REGS[i.dest].tag=y;
		REGS[i.dest].valid=0;
		ROB.dest_w(y,i.dest);
      }
      else
      {
         if(REGS[i.src1].valid)
         {
            ROB.data_w(y,REGS[i.src1].data);
			ROB.data_v_w(y,1);
            ROB.tag_w(y,-1);
         }
		 else if(ROB.ready_r(REGS[i.src1].tag))
         {
            ROB.data_w(y,ROB.data_r(REGS[i.src1].tag));
            ROB.tag_w(y,-1);
         }
         else if(CDB.checkTag(REGS[i.src1].tag))
         {
            ROB.data_w(y,CDB.read(REGS[i.src1].tag));
            ROB.tag_w(y,-1);
		 }
         else
         {
			ROB.tag_w(y,REGS[i.src1].tag);
         }
      }
      
      tail++;
      count++;

      if(tail==LSQSIZE)
       tail=0;

	  return 0;
   }
   
   int getReadyRS(void)
   {      
      if(!empty())
	  {
		  if(RS[head].opd1.valid)
			  return head;
	  }
	  return -1;
   }
   

/*   
   void beginLDex1(int x)
   {
      RS[x].imm = RS[x].imm + RS[x].opd1.data;
      RS[x].EAvalid = 1;
   }
   
   void beginLDex2(int x)
   {
      if(checkSCache(RS[x].imm))
         memAddr = RS[x].imm;
   }
   
   void beginExSW(int x)
   {
      ROB.R[RS[x].tag].dest = RS[x].imm + RS[x].opd1.data;
   }
   
   void exMemFetch(void)
   {
      if(inValid)
         outValid=1;
   }
*/
   
   void freeRS(int x)
   {
      RS[x].full=0;
      RS[x].tag=-1;
      if(head==(LSQSIZE-1))
         head=0;
      else
         head++;
      count--;
   }
   
   void check4Stall(void)
   {
      if(reqCDB==1)
         stall=1;
   }
/*   
   void execute(void)
   {
      int readyRS=getReadyRS();
      if(readyRS!=-1)
      {
         if(RS[readyRS].op==LD)
         {
            if(!RS[readyRS].EAvalid)
               BeginLDex1(readyRS);
            else
            {
               check4Stall();
               if(!stall)
               {
                  BeginLDex2(readyRS);
                  freeRS(readyRS);
               }
            }
         }
         else
         {
            beginExSW(readyRS);
            freeRS(readyRS);
         }
      }
   }
*/      
   void execute(void)
	{
	check4Stall();
	int readyRS=getReadyRS();
	printf("\n********LOAD/STORE EXECUTION***********");
	if(!stall)
	{
		mem.completeReq();
		checkOp();
		completeEx();
		beginEx(readyRS);
	}
	else
	{
		if(readyRS!=-1)
		{
			printf("\nPENDING EXECUTION: ");
			printInstrInfo(RS[readyRS].ip);
		}
	}
	displayEx();
	printf("\n************************************");
	}

void beginEx(int x)
{
	if(!stall)
	{
		if(x!=-1)
		{
		printf("\nEXECUTING: ");
		printInstrInfo(RS[x].ip);
		tagIn = RS[x].tag;
		ipIn = RS[x].ip;
		immIn =  RS[x].imm;
		opd1In = RS[x].opd1.data;
		opIn = RS[x].op;
		freeRS(x);
		}
		else
		{
		tagIn = -1;
		ipIn = -1;
		}
		
	}
}

void completeEx(void)
{
	if(!stall)
	{
		if(tagIn!=-1)
		{
			if(opIn==STORE)
			{
				ROB.dest_w(tagIn, 0);
				ROB.dest_v_w(tagIn,1);
				if(ROB.data_v_r(tagIn)==1)
					ROB.ready_w(tagIn, 1);
				tagOut1 = -1;
				ipOut1 = -1;
				opOut1 = -1;
			}//end if
			else if(opIn==LOAD)
			{
				tagOut1 = tagIn;
				ipOut1 = ipIn;
				opOut1 = opIn;
			}//end else if
		}//end if
		else
		{
			tagOut1 = -1;
			ipOut1 = -1;
			opOut1 = -1;
		}//end else
	}//end if
}

void checkOp(void)
{
	if(tagOut1!=-1)
	{
		if(opOut1==LOAD)
		{
			if(mem.free())
				mem.takeReq(tagOut1,ipOut1, !trace[ipOut1].loadMiss);
			else
				stall=1;
		}
	}
}

   void displayEx()
   {
	   printf("\nTag=%d:Valid=%d:Ack=%d:Stall=%d",tagOut,outValid,ack,stall);
   }

   void writeback(void)
   {
      if(ack)
      {
		 printf("\nResult on CDB ");
		 printInstrInfo(ROB.IP_r(tagOut));
         CDB.write(result,ack);
         CDB.updateTag(tagOut,ack);
         outValid=0;             // is this the proper place to do this?
//         ROB.ready_w(CDB.readTag(ack),1);
//         ROB.data_w(CDB.readTag(ack),CDB.read(CDB.readTag(ack)));
         ack=0;
		 reqCDB=0;
         stall=0;
      }
	  //else
	//	   CDB.floatTag();

	  if(reqCDB==1)
	  {
			printf("\n!!!!!!!!!!LSQ EXECUTE STALL CYCLE!!!!!!!!!!!!");
			printf("\nWAITING FOR CDB: ");
			printInstrInfo(ROB.IP_r(tagOut));
			//multUnitstallcycles++;
	  }
   }
   
   void updateRS(void)
   {
     for(int i=0; i<LSQSIZE; i++)
     {
      if(RS[i].full)
      {
         for(int j=1;j<=NUMCDB;j++)
         {
            if(!RS[i].opd1.valid && RS[i].opd1.tag==CDB.readTag(j))
            {
			   printf("\nUpdating LSQ RS[%d]: (OPD1) ",i);
			   printInstrInfo(RS[i].ip);
               RS[i].opd1.data=CDB.read(CDB.readTag(j));
               RS[i].opd1.valid=1;
            }
         }
      }
     }
     
   }

   	void displayRS(void)
    {
	   int i=head;
	   int n=count;

	   printf("\n*******LSQ RS CONTENTS*******");
	   while(n>0)
	   {
		   if(RS[i].opd1.valid)
		   {
			   printf("\n%d:READY ",i);
			   printInstrInfo(RS[i].ip);
		   }
		   else
		   {
			   printf("\n%d:WAITING ",i);
			   printInstrInfo(RS[i].ip);
		   }
			n--;
      
			if(i==LSQSIZE-1)
				i=0;
			else
				i++;
	   }
	   if(count==0)
		   printf("\n!!!!!!!!!!{EMPTY}!!!!!!!!!!");
	   printf("\n*******************************");
   }

   void reset(void)
   {
	  stall=0;
      ack=0;
      outValid=0;
	  tagOut=-1;
	  reqCDB=0;
	  head=0;
	  tail=0;
	  count=0;
	  opIn=-1;
	  opd1In=0;
	  result=0;
      tagOut=-1;
      tagIn=-1;
      outValid=0;

      opOut1=-1; 
	  ipOut1=-1; 
	  tagOut1=-1;
	  immIn=0;
	  ipIn=-1;
      
      for(int i=0;i<LSQSIZE;i++)
      {
        RS[i].tag=-1;
      }
   }

}LSQ;




	void memoryUnit::takeReq(int tagIn, int ipIn, int Hit)
	{
		tag=tagIn;
		ip=ipIn;
		hit=Hit;
		if(hit)
		{
			LSQ.reqCDB=1;
			responseTime=1;
			LSQ.tagOut=tag;
		}
		else
		{
			responseTime =MISSLATENCY-1;
			
		}
		busy=1;
	}

	void memoryUnit::completeReq(void)
	{
		if(busy)
		{
			responseTime--;
			if(responseTime==1)
			{
				LSQ.reqCDB=1;
				LSQ.tagOut=tag;
			}
			if(!responseTime)
				busy=0;
		}
	}



class branchUnit
{
public:
   int op;
   int opd1, opd2;
   int result;
   int tagOut;
   int outValid;
   int stall;
   int ack;
   int reqCDB;
   int brOutcome;
   RSslotType1 *RS;//[MAXRSBU];
   int *RSocc;
   int head;
   int tail;
   int count;  
   
   int branchUnitAlloc()
   {
	   if((RS=(RSslotType1*)calloc(NUMRSBU,sizeof(RSslotType1)))==NULL)
	   {
		   printf("\nNot enough memory to allocate BU RS data array");
		   return 0;
	   }
	   if((RSocc=(int*)calloc(NUMRSBU+1,sizeof(int)))==NULL)
	   {
		   printf("\nNot enough memory to allocate BU RS occupancy array");
		   return 0;
	   }
	  stall=0;
      ack=0;
      outValid=0;
	  tagOut=-1;
	  reqCDB=0;
	  head=0;
	  tail=0;
	  count=0;
	  result=0;
      outValid=0;

      for(int i=0;i<NUMRSBU;i++)
      {
        RS[i].tag=-1;
      }
	  return 1;
   }

    void calcUtil(void)
	{
		RSocc[count]++;
	}

   	int full(void)
	{
		if(count==NUMRSBU)
			return 1;
		else
			return 0;
	}

	int empty(void)
	{
		if(count==0)
			return 1;
		else
			return 0;
	}

   int getFreeRS(void)
   {
     if(!full())
      return tail;
     else
      return -1;
   }
   
   int issue(traceElement i)
   {
	  int y;
      brOutcome = i.bOutcome;
      int x=getFreeRS();
      if(x==-1)
         return -1;

	  y=ROB.enqueue(i.opcode,i.addr);
	  ROB.ready_w(y,0);
      RS[x].op=i.opcode;
      RS[x].full=1;
      RS[x].tag=y;
	  RS[x].ip=i.addr;      
      
      //for instruction i source operand 1
      if(REGS[i.src1].valid)
      {
         RS[x].opd1.valid=1;
         RS[x].opd1.data=REGS[i.src1].data;
      }
      else if(CDB.checkTag(REGS[i.src1].tag))
      {
         RS[x].opd1.valid=1;
         RS[x].opd1.data=CDB.read(REGS[i.src1].tag);
      }
      else if(ROB.ready_r(REGS[i.src1].tag))
      {
         RS[x].opd1.valid=1;
         RS[x].opd1.data=ROB.data_r(REGS[i.src1].tag);
      }
      else
      {
         RS[x].opd1.valid=0;
         RS[x].opd1.tag=REGS[i.src1].tag;
      }
      
      if(i.opcode==BNE)
      {
         //for instruction i source operand 2
         if(REGS[i.src2].valid)
         {
            RS[x].opd2.valid=1;
            RS[x].opd2.data=REGS[i.src2].data;
         }
         else if(CDB.checkTag(REGS[i.src2].tag))
         {
            RS[x].opd2.valid=1;
            RS[x].opd1.data=CDB.read(REGS[i.src2].tag);
         }
         else if(ROB.ready_r(REGS[i.src2].tag))
         {
            RS[x].opd2.valid=1;
            RS[x].opd1.data=ROB.data_r(REGS[i.src2].tag);
         }
         else
         {
            RS[x].opd2.valid=0;
            RS[x].opd2.tag=REGS[i.src2].tag;
         }
      }
      else
      {
         RS[x].opd2.valid=1;
      }
       
      ROB.dest_w(y,i.bPredict);

      
      tail++;
      count++;

      if(tail==NUMRSBU)
       tail=0;

	  return 0;
   }
   
   int getReadyRS(void)
   {      
      if(!empty())
	  {
		if(RS[head].opd1.valid && RS[head].opd2.valid)
         return head;
      }
         return -1;
   }
   
   void freeRS(int x)
   {
      RS[x].full=0;
      RS[x].tag=-1;
      if(head==(NUMRSBU-1))
         head=0;
      else
         head++;
      count--;
   }
   
   void check4Stall(void)
   {
      if(reqCDB==1)
         stall=1;
   }

   void BeginEx(int x)
   {
      outValid=1;   // who will reset this? the arbiter?
      result=trace[RS[x].ip].bOutcome;
      tagOut=RS[x].tag;
	  reqCDB=1;
   }
   
   
   void execute(void)
   {
      check4Stall();
	  int readyRS=getReadyRS();
	  printf("\n********BU EXECUTION***********");
      if(!stall)
      {
         if(readyRS!=-1)
         {
			printf("\nEXECUTING: ");
			printInstrInfo(RS[readyRS].ip);
            BeginEx(readyRS);
            freeRS(readyRS);
         }
      }
	  else
	  {
			if(readyRS!=-1)
			{
				//printf("\n!!!!!!!!!!INT ALU EXECUTE STALL CYCLE!!!!!!!!!!!!");
				printf("\nPENDING EXECUTION: ");
				printInstrInfo(RS[readyRS].ip);
				//intALUstallcycles++;
			}
			
	  }
	  displayEx();
	  printf("\n************************************");
   }
   
   void displayEx()
   {
	   printf("\nTag=%d:Valid=%d:Ack=%d:Stall=%d",tagOut,outValid,ack,stall);
   }

   void writeback(void)
   {
      if(ack)
      {
		 printf("\nResult on CDB ");
		 printInstrInfo(ROB.IP_r(tagOut));
         CDB.write(result,ack);
         CDB.updateTag(tagOut,ack);
         outValid=0;             // is this the proper place to do this?
//         ROB.ready_w(CDB.readTag(ack),1);
//         ROB.data_w(CDB.readTag(ack),CDB.read(CDB.readTag(ack)));
         ack=0;
		 reqCDB=0;
         stall=0;
      }
	  //else
	  //	   CDB.floatTag();

	  if(reqCDB==1)
	  {
			printf("\n!!!!!!!!!!BU EXECUTE STALL CYCLE!!!!!!!!!!!!");
			printf("\nWAITING FOR CDB: ");
			printInstrInfo(ROB.IP_r(tagOut));
			intALUstallcycles++;
	  }

      //snooping function
      //updateRS();
   }
   
   void updateRS(void)
   {
     for(int i=0; i<NUMRSBU; i++)
     {
      if(RS[i].full)
      {
         for(int j=1;j<=NUMCDB;j++)
         {
            if(!RS[i].opd1.valid && RS[i].opd1.tag==CDB.readTag(j))
            {
			   printf("\nUpdating BU RS[%d]: (OPD1) ",i);
			   printInstrInfo(RS[i].ip);
               RS[i].opd1.data=CDB.read(CDB.readTag(j));
               RS[i].opd1.valid=1;
            }
            if(!RS[i].opd2.valid && RS[i].opd2.tag==CDB.readTag(j))
            {
			   printf("\nUpdating BU RS[%d]: (OPD2) ",i);
			   printInstrInfo(RS[i].ip);
               RS[i].opd2.data=CDB.read(CDB.readTag(j));
               RS[i].opd2.valid=1;
            }
         }
      }
     }
     
   }

    void displayRS(void)
    {
	   int i=head;
	   int n=count;

	   printf("\n*******BU RS CONTENTS*******");
	   while(n>0)
	   {
		   if(RS[i].opd1.valid && RS[i].opd2.valid)
		   {
			   printf("\n%d:READY ",i);
			   printInstrInfo(RS[i].ip);
		   }
		   else
		   {
			   printf("\n%d:WAITING ",i);
			   printInstrInfo(RS[i].ip);
			   if(!RS[i].opd1.valid)
					printf(" OPD1 ");
			   if(!RS[i].opd2.valid)
					printf(" OPD2 ");
		   }
			n--;
      
		   if(i==NUMRSBU-1)
				i=0;
		   else
				i++;
	   }
	   if(count==0)
		   printf("\n!!!!!!!!!!{EMPTY}!!!!!!!!!!");
	   printf("\n*******************************");
   }

   void reset(void)
   {
	  stall=0;
      ack=0;
      outValid=0;
	  tagOut=-1;
	  reqCDB=0;
	  head=0;
	  tail=0;
	  count=0;
	  result=0;
      tagOut=-1;
      outValid=0;

      
      for(int i=0;i<NUMRSBU;i++)
      {
        RS[i].tag=-1;
      }
   }
}BRunit;




//////////////////////////////////////////////////////////////////////////////
class sCache
{
public:
   int *S;//[SCACHESIZE];
   
   int sCacheAlloc()
   {
	   if((S=(int*)malloc(sizeof(int)*SCACHESIZE))==NULL)
	   {
		   printf("\nNot enough memory to allocate SCACHE data array");
		   return 0;
	   }
      for(int i=0;i<SCACHESIZE;i++)
         S[i]=-1;
	  return 1;
   }
  
   void deleteAddr(int x)
   {
      for(int i=0;i<SCACHESIZE;i++)
         if(S[i]==x)
            S[i]=-1;
   }
   
   void addAddr(int x)
   {
      for(int i=0;i<SCACHESIZE;i++)
      {
         if(S[i]==-1)
         {   
            S[i]=x;
            break;
         }
      }   
   }
   
   void display()
   {
	   printf("\nSW CACHE CONTENTS");
	   for(int j=0;j<SCACHESIZE;j++)
		   printf("\n%d:%d",j,S[j]);
   }
   
};
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
sCache SCache;
//////////////////////////////////////////////////////////////////////////////


void commonDataBus::arbiter(int vBRU,int vDIV,int vLSQ,int vMULT,int vFPADD,int vALU)
   {
      int r1,r2,r3,r4,r5,r6;
      r1=r2=r3=r4=r5=r6=1;
	  static int delay=3;
      
      for(int j=0;j<NUMCDB;j++)
      {
         if(vBRU&&r1)
         {  
            BRunit.ack=j+1;
            r1=0;
         }
         else if(vDIV&&r2)
         {  
        //    DIVunit.ack=j+1;
            r2=0;
         }         
         else if(vLSQ&&r3)
         {  
            LSQ.ack=j+1;
            r3=0;
         }         
         else if(vMULT&&r4)
         {  
            multUnit.ack=j+1;
            r4=0;
         }                  
         else if(vFPADD&&r5)
         {  
        //    FPADDunit.ack=j+1;
            r5=0;
         }         
         else if(vALU&&r6)
         {  
		    //if(!delay)
			//{
				intALU.ack=j+1;
				//delay=3;
			//}
			//else
			//	delay--;
            r6=0;
         }
		 else
			 floatTag(j);
         
      }
   }


//////////////////////////////////////////////////////////////////////////////
class fetchBuffer
{
public:
   traceElement *buffer;//[ISSUESIZE];
   
   int head;
   int tail;
   int count;
   int lastIndex;
   
   int fetchBufferAlloc()
   {
	   if((buffer=(traceElement*)malloc(sizeof(traceElement)*ISSUESIZE))==NULL)
	   {
		   printf("\nNot enough memory to allocate fetch buffer data array");
		   return 0;
	   }
      head=0;
      tail=0;
      count=0;
      lastIndex=0;
	  return 0;
   }
   
   int full()
   {
      if(count==ISSUESIZE)
         return 1;
      else
         return 0;      
   }
   
   int empty()
   {
      if(count==0)
         return 1;
      else
         return 0;      
   }
   
   int enqueue(traceElement i)
   {
      if(full())
         return -1;
      buffer[tail].addr=i.addr;
      buffer[tail].opcode=i.opcode;
      buffer[tail].dest=i.dest;
      buffer[tail].src1=i.src1;
      buffer[tail].src2=i.src2;
      buffer[tail].imm=i.imm;
      buffer[tail].opd1=i.opd1;
      buffer[tail].opd2=i.opd2;
      buffer[tail].branchDest=i.branchDest;
      buffer[tail].takenCount=i.takenCount;
      buffer[tail].bPredict=i.bPredict;
      buffer[tail].bOutcome=i.bOutcome;
      buffer[tail].loadMiss=i.loadMiss;
      
      count++;
      lastIndex=tail;
      tail++;
      if(tail==ISSUESIZE)
       tail=0;      
      return lastIndex;      
   }
   
   int dequeue()
   {
      if(empty())
         return -1;
      
      count--;
      
      if(head==ISSUESIZE-1)
       head=0;
      else
       head++;
      
      return head;      
   }
   
   traceElement readElement()
   {
      return buffer[head];
   }

   void display()
   {
	   int i=head;
	   int n=count;
	   printf("\n*****IQ CONTENTS*****");
	   if(count==0)
		   printf("\n!!!!!! {EMPTY} !!!!!!");
	   else
	   {
		while(n>0)
		{
		   	printf("\n");
			printInstrInfo(buffer[i].addr);
			if(i==ISSUESIZE-1)
				i=0;
			else
				i++;
			n--;
		}
	   }
	   printf("\nTotal Queued:%d",count);
	   printf("\n*********************");
   }
   
};
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
fetchBuffer IQ;
//////////////////////////////////////////////////////////////////////////////



//////////////////////////////////////////////////////////////////////////////
class missPredictStack
{
public:
   int top;
   char *S;//[STACKSIZE];

   int missPredictStackAlloc()
   {
	   if((S=(char*)malloc(sizeof(char)*STACKSIZE))==NULL)
	   {
		   printf("\nNot enough memory to allocate stack buffer data array");
		   return 0;
	   }

      top=-1;
	  return 1;
   }
   
   int full()
   {
      if(top==STACKSIZE-1)
         return(1);
      else
         return(0);
   }

   int empty()
   {
      if(top==-1)
         return(1);
      else
         return(0);
   }

   void push(int a)
   {
      top++;
      S[top]=a;
   }

   int pop()
   {
      return(S[top--]);
   }

};
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
missPredictStack pcStack;
//////////////////////////////////////////////////////////////////////////////



//////////////////////////////////////////////////////////////////////////////
int main(int argc, char *argv[])
{
    int cycles=0;
    int brMissPredict=0;
    char file[]="prog4.txt";
    
	if(!readConfigFile())
      exit(1);

    //initialize datastructures here
   CDB.commonDataBusInit();
   ROB.reOrderBufferAlloc();
   intALU.simpleALUAlloc();
   multUnit.pipelinedMultiplierAlloc();
   LSQ.loadStoreQueueAlloc();
   IQ.fetchBufferAlloc();
   SCache.sCacheAlloc();
   pcStack.missPredictStackAlloc();
   BRunit.branchUnitAlloc();

   if(!readTraceFile(argv[1]))
      return(0);

    initializeREGS();
    for(;;)
    { 
	  printf("\n**********SIMULATION CYCLE %d OUTPUT***********",cycles+1);
	  //intALU.displayRS();
	  ///////////////////////////COMMIT()//////////////////////////////
      brMissPredict=commit();
	  ////////////////////////////UPDATES/////////////////////////////
	  updateReservationStations();
	  ROB.updateROB();
	  ROB.display();
	  intALU.displayRS();
	  multUnit.displayRS();
	  LSQ.displayRS();
	  BRunit.displayRS();
	  /////////////////////////ARBITER()//////////////////////////////
	  //CDB.arbiter(BRunit.outValid,DIVunit.outValid,LSQ.outValid,
      //            MULTunit.outValid,FPADDunit.outValid,intALU.outValid);
	  CDB.arbiter(BRunit.reqCDB,0,LSQ.reqCDB,multUnit.reqCDB,0,intALU.reqCDB);
	  
	  ///////////////////////////WRITEBACK()///////////////////////////
	  printf("\n*********CDB Status**********");
      writeback();
	  CDB.display();
	  //intALU.displayRS();
	  /////////////////////////EXECUTE()//////////////////////////////
      execute();
	  //intALU.displayEx();
	  //IQ.display();
	  /////////////////////////ISSUE()//////////////////////////////
      issue();
	  intALU.calcUtil();
	  multUnit.calcUtil();
	  LSQ.calcUtil();
	  BRunit.calcUtil();
	  ROB.calcUtil();
	  //intALU.displayRS();
	  /////////////////////////FETCH()//////////////////////////////
      fetch(brMissPredict);
	  IQ.display();
	  cycles++;
	  printf("\n");
      if(ROB.empty() && IQ.empty())
         break;
	  getchar();
    }

	printf("\nTotal Instructions Executed: %d",IC);
	printf("\nTotal Fetch Stall Cycles: %d",fetchstallcycles);
	printf("\nTotal Issue Stall Cycles: %d",fetchstallcycles);
	printf("\nTotal INT ALU Stall Cycles: %d",intALUstallcycles);
	printf("\nTotal MULT Unit Stall Cycles: %d",multUnitstallcycles);
	printf("\nTotal Cycles: %d",cycles);
	printf("\nCycles per Instruction: %.2f",(double)cycles/(double)IC);
	printf("\nINT ALU Utilization: %.0f %%",(double)intALUutil/(double)cycles*100);
	printf("\nMULT UNIT Utilization: %.0f %%",(double)multUnitutil/(double)cycles*100);
	for(int i=0;i<NUMCDB;i++)
		printf("\nCDB %d utilization: %.0f %%",i,(double)CDB.util[i]/(double)cycles*100);
	outputUtils();

	return(0);
}
//////////////////////////////////////////////////////////////////////////////

void outputUtils(void)
{
	FILE *file;

	if((file=fopen("utils.txt","w"))==NULL)
	{
		printf("\nCan't create file for utilization data output!!");
		return;
	}

	fprintf(file,"\nStats for #issues for Issue Size %d",ISSUESIZE);
	for(int i=0;i<ISSUESIZE+1;i++)
	{
		fprintf(file,"\n%d",numissued[i]);
	}

	fprintf(file,"\nStats for #commits for Commit Size %d",COMMITSIZE);
	for(i=0;i<COMMITSIZE+1;i++)
	{
		fprintf(file,"\n%d",numcommitted[i]);
	}

	fprintf(file,"\nStats for ROB occupancy for ROB size %d",ROBSIZE);
	for(i=0;i<ROBSIZE+1;i++)
	{
		fprintf(file,"\n%d",ROB.ROBocc[i]);
	}

	fprintf(file,"\nStats for INT ALU RS occupancy for #RS=%d ",NUMRSINTALU);
	for(i=0;i<NUMRSINTALU+1;i++)
	{
		fprintf(file,"\n%d",intALU.RSocc[i]);
	}
	fclose(file);
}
//////////////////////////////////////////////////////////////////////////////
void assignValue(int n,int a,int b,int c,int d,int e,int f,int g,int h,int i,int j,int k,int l)
{
	trace[n].opcode=a;
	trace[n].dest=b;
	trace[n].src1=c;
	trace[n].src2=d;
	//trace[n].opcode=e;
}

int readConfigFile(void)
{
   FILE *file;
   char line[80];

   if((file=fopen("params.txt","r"))==NULL)
   {
      fprintf(stderr,"\nWARNING: Can't open configurable parameters file.");
      return 0;
   }

   //fgets(line, 80, file);
   //NUMINTREGS=atoi(line);

   //fgets(line, 80, file);
   //NUMFPREGS=atoi(line);

   fgets(line, 80, file);
   ISSUESIZE=parseParam(line); 

   fgets(line, 80, file);
   COMMITSIZE=parseParam(line);

   fgets(line, 80, file);
   ROBSIZE=parseParam(line);       

   fgets(line, 80, file);
   NUMCDB=parseParam(line); 

   fgets(line, 80, file);
   LSQSIZE=parseParam(line);

   fgets(line, 80, file);
   SCACHESIZE=parseParam(line);

   fgets(line, 80, file);
   NUMRSBU=parseParam(line);

   fgets(line, 80, file);
   NUMRSINTALU=parseParam(line);

   fgets(line, 80, file);
   NUMRSMULT=parseParam(line);

   fgets(line, 80, file);
   MULTSTAGES=parseParam(line);

   fgets(line, 80, file);
   NUMRSDIV=parseParam(line);

   fgets(line, 80, file);
   DIVCYCLES=parseParam(line);

   fgets(line, 80, file);
   NUMRSFPADD=parseParam(line);

   fgets(line, 80, file);
   FPADDSTAGES=parseParam(line);  
   
   fgets(line, 80, file);
   MISSPROB=parseParam(line);
   
   fgets(line, 80, file);
   MPPROB=parseParam(line);
   
   fclose(file);
   
   return 1;
}

int parseParam(char *s)
{
	while(*s != '=')
		s++;
	s++;
	return(atoi(s));
}
//////////////////////////////////////////////////////////////////////////////
int readTraceFile(char *traceFile)
{
   FILE *file;
   char line[200];
   int tokens[NUMTOKENS];
   int i=0;
   
   //gets(line);
   file=fopen(traceFile,"r");

   if(file==NULL)
   {
      printf("\nWARNING: Can't open specified program trace file.");
	  getchar();
      return 0;
   }
   
   while(!feof(file))
   {
      fgets(line,200,file);
      parseLine(line, tokens);
      readTokens(tokens,i);
	  trace[i].addr=i;
      i++;
   }
   return i;
}
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////      
void parseLine(char *line, int *tokens)
{
   int i=0,j;
   char token[20];
   int count=0;
   
   while(count<NUMTOKENS)
   {
      j=0;
      while(line[i]!=',' && line[i]!=';')
         token[j++]=line[i++];
      token[j]='\0';
      i++;
      tokens[count]=atoi(token);
	  count++;
   }
}      
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
void readTokens(int *tokens, int i)
{     
   trace[i].opcode=tokens[0];
   trace[i].dest=tokens[1];
   trace[i].src1=tokens[2];
   trace[i].src2=tokens[3];
   trace[i].imm=tokens[4];
   trace[i].opd1=tokens[5];
   trace[i].opd2=tokens[6];
   trace[i].branchDest=tokens[7];
   trace[i].takenCount=tokens[8];
   trace[i].bPredict=tokens[9];
   trace[i].bOutcome=tokens[10];
   trace[i].loadMiss=tokens[11];
}
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
void initializeREGS()
{
	for(int i=0;i<INTREGS+FPREGS;i++)
	{
		REGS[i].data=0;
		REGS[i].valid=0;
		REGS[i].tag=-1;
	}
	REGS[1].valid=1;
	REGS[2].valid=1;
	REGS[34].valid=1;
	//REGS[8].valid=1;
	//REGS[9].valid=1;
	//REGS[11].valid=1;
	//REGS[12].valid=1;
	//REGS[15].valid=1;
	//REGS[16].valid=1;
	//REGS[17].valid=1;
}
//////////////////////////////////////////////////////////////////////////////



//////////////////////////////////////////////////////////////////////////////
int commit(void)
{
   int i;
   int writeReg=-1;
   int memoryWriteAddr=-1;
   int x=0;
   int committed=0;

	
   printf("\nCommitting instructions...");

   for(i=0;i<COMMITSIZE;i++)
   {
      if(!ROB.empty())
      {
         x=ROB.head;
         if(ROB.ready_r(x)==1)
         {
            if(ROB.instrType_r(x)==BRANCH)
            {
               //if(ROB.dest_r(x)!=ROB.data_r(x))
               //{
                //  resetState();
				 // printf("\nBranch Mispredict Occured. State Reset!");
                  //return 1;
               //}
            }
            else if(ROB.instrType_r(x)==STORE)
            {
               memoryWriteAddr = ROB.dest_r(x);
               SCache.deleteAddr(memoryWriteAddr);
            }
            else
            {
               writeReg=ROB.dest_r(x);
               REGS[writeReg].data=ROB.data_r(x);
               if(x==REGS[writeReg].tag)
                  REGS[writeReg].valid = 1;
            }
			printf("\nCOMMITTED: ");
			printInstrInfo(ROB.IP_r(x));
			committed++;
            ROB.dequeue();
         }
      }
      else
         break;
   }
   numcommitted[committed]++;
   printf("\nTotal Committed: %d/%d",committed,COMMITSIZE);
   return 0;
}
//////////////////////////////////////////////////////////////////////////////
void printInstrInfo(int IP)
{
	int instrType;
	if(IP==-1)
		return;
	int Opcode=trace[IP].opcode;
	 if(Opcode==BNEZ || Opcode==BNE)
         instrType = BRANCH;
      else if(Opcode==SW)
         instrType = STORE;
      else if(Opcode==LW)
         instrType = LOAD;
      else
         instrType = ALU;
	if(instrType==BRANCH)
    {
		if(trace[IP].opcode==BNEZ)
			printf("BNEZ R%d, %d\n",trace[IP].src1,trace[IP].branchDest);
		else
			printf("BNE R%d, R%d, %d",trace[IP].src1,trace[IP].src2,trace[IP].branchDest);
	}
    else if(instrType==STORE)
		printf("SW R%d, %d(R%d)",trace[IP].src1,trace[IP].imm,trace[IP].src2);
    else if(instrType==LOAD)
		printf("LW R%d, %d(R%d)",trace[IP].dest,trace[IP].imm,trace[IP].src1);
    else
	{
		switch(trace[IP].opcode)
		{
		case ADD:
		printf("ADD R%d, R%d, R%d",trace[IP].dest,trace[IP].src1,trace[IP].src2);
		break;
		case FADD:
		printf("FADD F%d, F%d, F%d",trace[IP].dest-FPREGS,trace[IP].src1-FPREGS,trace[IP].src2-FPREGS);
		break;
		case SUB:
		printf("SUB R%d, R%d, R%d",trace[IP].dest,trace[IP].src1,trace[IP].src2);
		break;
		case FSUB:
		printf("FSUB F%d, F%d, F%d",trace[IP].dest-FPREGS,trace[IP].src1-FPREGS,trace[IP].src2-FPREGS);
		break;
		case MUL:
		printf("MUL R%d, R%d, R%d",trace[IP].dest,trace[IP].src1,trace[IP].src2);
		break;
		case DIV:
		printf("DIV R%d, R%d, R%d",trace[IP].dest,trace[IP].src1,trace[IP].src2);
		break;
		case FDIV:
		printf("FDIV F%d, F%d, F%d",trace[IP].dest-FPREGS,trace[IP].src1-FPREGS,trace[IP].src2-FPREGS);
		break;
		case FMUL:
		printf("FMUL F%d, F%d, F%d",trace[IP].dest-FPREGS,trace[IP].src1-FPREGS,trace[IP].src2-FPREGS);
		break;
		case DADDUI:
		printf("DADDUI R%d, R%d, R%d",trace[IP].dest,trace[IP].src1,trace[IP].src2);
		break;
		}
	}
}


//////////////////////////////////////////////////////////////////////////////
void resetState(void)
{
   ROB.reset();
   intALU.reset();
   multUnit.reset();
   //FPADDunit.reset();
   //DIVunit.reset();
   BRunit.reset();
   LSQ.reset();
}
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
void updateReservationStations(void)
{   
   intALU.updateRS();
   multUnit.updateRS();
   //DIVunit.updateRS();
   //FPADDunit.updateRS();
   BRunit.updateRS();
   LSQ.updateRS();
}
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
void writeback(void)
{
   intALU.writeback();
   multUnit.writeback();
   //DIVunit.writeback();
   //FPADDunit.writeback();
   BRunit.writeback();
   LSQ.writeback();
}
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
void execute(void)
{
   intALU.execute();
   multUnit.execute();
   //DIVunit.execute();
   //FPADDunit.execute();
   BRunit.execute();
   LSQ.execute();
}
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
void issue(void)
{
    traceElement i;	//to hold current instruction
    int RSfull=0;   //to check when to stop issue
	int issued=0;   //total instructions issued


	printf("\nIssuing instructions....");

    while(!IQ.empty() && !ROB.full()) //while instructions in queue and 
									  //ROB not full
    {
		i=IQ.readElement();	// get a instruction from the queue
            
		 switch(i.opcode)   //check the opcode to decide which RS to issue to
		 {
		    case DADDUI:
			case ADD:
            case SUB:
				RSfull=intALU.issue(i);
				if(RSfull)
					printf("\nINTEGER ALU RS are full!!!");
                break;
            case MUL:
            case FMUL:
                RSfull=multUnit.issue(i);
				if(RSfull)
					printf("\nMULT UNIT RS are full!!!");
                break;
            case DIV:
            case FDIV:
                //RSfull=DIVunit.issue(i);
                break;
            case BNEZ:
            case BNE:
                RSfull=BRunit.issue(i);
				if(RSfull)
					printf("\nBU UNIT RS are full!!!");
                break;
            case SW:
            case LW:
                RSfull=LSQ.issue(i);
				if(RSfull)
					printf("\nLSQ RS are full!!!");
                break;            
		}
		if(RSfull) //if the RS was found to be full, stop issue
			break;
        else	   //else remove from queue
		{
			IQ.dequeue();
			printf("\nISSUED: ");
			printInstrInfo(i.addr);
			issued++;
		}
	}
	if(!IQ.empty() && issued==0)
	{
		printf("\n!!!!!!!!!!ISSUE STALL CYCLE!!!!!!!!!!!!");
		issuestallcycles++;
	}
	numissued[issued]++;
	printf("\nTotal Issued:%d/%d",issued,ISSUESIZE);
}
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
void fetch(int branchMiss)
{
   static int PC; //The instruction counter.
				  //Defined as a static variable to hold the previous state.
   traceElement t; //To hold current read instruction from the trace
   int fetched=0;  //To hold number of instructions fetched
   static int EOP;
   
   //now we're deciding upon the current PC to start fetching from.
   //if there has been a branch miss-predict, fetch from the target 
   //of that miss predicted branch

   printf("\nFetching new instructions...");
   if(branchMiss)
   {
	printf("\nA branch miss detected!");
	PC=pcStack.pop();
   }
   //else 
   //continue fetching after the last fetched instruction. 
   //PC already points there.

   
   while(!IQ.full()) //while capacity remains according to FETCHSIZE										 
   {
      t=trace[PC];	//fetch the instruction
      //examine what is type of current instruction.  
	  //We're interested in BRANCH, LOAD and EXIT instructions here for 
	  //special handling of these.
      
      if(t.opcode==BNEZ || t.opcode==BNE)	//if BRANCH
      {         
		 //get a branch prediction for this instruction
         t.bPredict=getBranchPrediction();	
         //examine the taken count to see if actually it should be taken
		 //NOTE: this feature applies to loop branches mainly that reoccur.
         if(t.takenCount>0) //if this is a taken branch(actually)
         {
            t.bOutcome=1;
            trace[PC].takenCount--;	//decrement taken count(useful for loops)
            //if(t.bOutcome==t.bPredict)	//if the prediction is matching
               PC = PC + t.branchDest;  //goto branch destination
            //else
            //{
              // pcStack.push(PC+t.branchDest); //else, misprediction happened.
											  //target PC saved on stack
               //PC = PC + 1; //just increment the PC
			   //printf("\nT branch miss predicted!");
           // }
         }
         else //if not a taken branch(actually)
         {
            t.bOutcome=0;
            //if(t.bOutcome==t.bPredict) //if the prediction is matching,
               PC = PC + 1;			   //just increment the PC.
            //else					   //else misprediction happened
            //{
               //pcStack.push(PC+1);	   //target PC saved on stack
              // PC = PC + t.branchDest; //goto branch destination
			  // printf("\nNT branch miss predicted!");
           // }
         }
      }
      else if(t.opcode==LW)  //else if LOAD
      {
         t.loadMiss=getLoadPrediction(); //get a load miss prediction
		 printf("\nLoad Miss Prediction=%d for ",t.loadMiss);
		 printInstrInfo(PC);
         PC = PC + 1;					 //just increment the PC
      }
      else if(t.opcode==EXIT) //else if EXIT
      {
		  EOP=1;
		 printf("\nNO MORE NEW INSTRUCTIONS TO FETCH!");
         break;				  //stop fetching anymore. No need to increment PC
	  }
      else
         PC = PC + 1;		  // else a normal intruction. Increment PC.
            
      IQ.enqueue(t); //queue for issue.
	  printf("\nPC=%d: ",(PC-1));
	  printInstrInfo(t.addr);
	  fetched++;
	  IC++;
   }
   printf("\nTotal Fetched: %d",fetched);
   if(fetched==0 && !EOP)
   {
	   fetchstallcycles++;
	   printf("\n!!!!!FETCH STALL CYCLE ENCOUNTERED!!!!!");
   }
}
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
int getBranchPrediction()
{
	int r=rand()%100+1;

	if(r>=MPPROB)
		return 1;
	else
		return 0;
}
//////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////
int getLoadPrediction()
{
	int r=rand()%100+1;

	if(r>=MISSPROB)
		return 1;	
	else
		return 0;
}